def create_table(table_name, fields):
    """
    String, Iterable --> String
    """
    sql = "CREATE TABLE {} (".format(table_name)
    sql += ', '.join(["{} {}".format(field, fields[field]) for field in fields])
    sql += ')'
    return sql


def insert_many(table_name, rows):
    '''
    Given a table name and a list of dictionaries representing
    rows, generate a (sql, template) tuple of strings that can be
    passed to psycopg2.extras.execute_values() [1] to bulk insert all the
    values for improved efficiency [2].

    For example:

        >>> insert_many('boop', [{'foo': 1, 'bar': 2}])
        ('INSERT INTO boop (foo, bar) VALUES %s', '(%(foo)s, %(bar)s)')

    [1]: http://initd.org/psycopg/docs/extras.html#psycopg2.extras.execute_values
    [2]: https://stackoverflow.com/a/30985541
    '''

    field_names = list(rows[0].keys())
    fields = ', '.join(field_names)
    placeholders = ', '.join(["%({})s".format(k) for k in field_names])
    template = f"({placeholders})"
    sql = f"INSERT INTO {table_name} ({fields}) VALUES %s"

    return sql, template



"""
Each function in this file is the name of a table or dataset.
"""

from .transform import with_geo, with_bbl, to_csv, extract_csvs_from_zip, extract_csv_from_zip, skip_fields
from .transform import hpd_registrations_address_cleanup, hpd_contacts_address_cleanup
from .dof_parser import parse_dof_file
from .datasets import datasets

def ecb_violations(dataset):
    return with_bbl(to_csv(dataset.files[0].dest), borough='boro')


def dob_violations(dataset):
    return with_bbl(to_csv(dataset.files[0].dest), borough='boro')


def _pluto(dataset):
    """
    Handles importing of all pluto versions,
    optionally allowing for skipped fields in some versions.
    It assumes there is only one schema for each pluto dataset.
    """
    pluto_generator = with_geo(to_csv(extract_csvs_from_zip(dataset.files[0].dest)))
    pluto_fields_to_skip = dataset.schemas[0].get('skip')

    if pluto_fields_to_skip:
        return skip_fields(pluto_generator, [s.lower() for s in pluto_fields_to_skip])

    return pluto_generator

# Creates a function for each pluto version
# same as doing def pluto_15v1() ... def pluto_16v2 ... etc
for pluto_version in filter(lambda x: x[0:5] == 'pluto', datasets().keys()):
    exec(f'''
def {pluto_version}(dataset):
    return _pluto(dataset)
''')

def hpd_complaints(dataset):
    return with_bbl(to_csv(dataset.files[0].dest))


def hpd_complaint_problems(dataset):
    return to_csv(dataset.files[1].dest)


def dob_complaints(dataset):
    return to_csv(dataset.files[0].dest)


def hpd_violations(dataset):
    return to_csv(dataset.files[0].dest)


def hpd_registrations(dataset):
    return hpd_registrations_address_cleanup(to_csv(dataset.files[0].dest))


def hpd_contacts(dataset):
    return hpd_contacts_address_cleanup(to_csv(dataset.files[1].dest))


def dof_sales(dataset):
    for f in dataset.files:
        for row in with_bbl(parse_dof_file(f.dest)):
            yield row


def dobjobs(dataset):
    return with_bbl(to_csv(dataset.files[0].dest))


def rentstab(dataset):
    return to_csv(dataset.files[0].dest)


def rentstab_v2(dataset):
    return to_csv(dataset.files[0].dest)


def rentstab_summary(dataset):
    return to_csv(dataset.files[0].dest)


def acris(dataset, schema):
    dest_file = next(filter(lambda f: schema['table_name'] in f.dest, dataset.files))
    _to_csv = to_csv(dest_file.dest)
    if 'skip' in schema:
        return skip_fields(_to_csv, [s.lower() for s in schema['skip']])
    else:
        return _to_csv


def oath_hearings(dataset):
    return with_bbl(to_csv(dataset.files[0].dest),
                    borough='violationlocationborough',
                    block='violationlocationblockno',
                    lot='violationlocationlotno')

def pad_adr(dataset):
    pad_generator = with_bbl(to_csv(extract_csv_from_zip(
        dataset.files[0].dest, 'bobaadr.txt')), borough='boro')

    pad_fields_to_skip = dataset.schemas[0].get('skip')

    return skip_fields(pad_generator, [s.lower() for s in pad_fields_to_skip])


def j51_exemptions(dataset):
    return with_bbl(to_csv(dataset.files[0].dest), borough='boroughcode')


def marshal_evictions(dataset, schema):
    dest_file = next(filter(lambda f: schema['table_name'] in f.dest, dataset.files))
    _to_csv = to_csv(dest_file.dest)
    return _to_csv


def hpd_vacateorders(dataset):
    return to_csv(dataset.files[0].dest)



"""
Functions to standardize address strings
in HPD contacts data
"""
import re

STREETS = [
    (r'(?<= )AVE(?= |$)', 'AVENUE'),
    (r'(?<= )(STREET|STR|(ST\.?))(?= |$)', 'STREET'),
    (r'(?<= )PL(?= |$)', 'PLACE'),
    (r'(?<= )(ROAD|(?<!\d)RD\.?)', 'ROAD'),
    (r'(?<= )(LA|LN)(?= |$)', 'LANE'),
    (r'(?<= )CT|CRT', 'COURT'),
    (r'(?<= )DR\.?(?= |$)', 'DRIVE'),
    (r'(?<= )(BOULEVARD|BLVD)', 'BOULEVARD'),
    (r'(?<= )(PKWY|PARKWY)', 'PARKWAY'),
    (r'(?<= )(PK)', 'PARK'),
    (r'(?<= )(BCH)', 'BEACH'),
    (r'(?<= )TERR(?= |$)', 'TERRACE'),
    (r'(^|(?<= ))(BDWAY|BDWY|BROAD WAY)', 'BROADWAY')
]


# Look for direction abbrivation as the start of the string or a space, but NOT 'AVENUE '
# this is to avoid lettered avenues such as "AVENUE W"
DIR_START = "(^|(?<=[ ]))(?<!AVENUE )"
DIR_END = "((?=[ ])|$)"

dir_regex = lambda x: DIR_START + x + DIR_END

DIRECTIONS = [
    (dir_regex(r'N\.?'), 'NORTH'),
    (dir_regex(r'SO?\.?'), 'SOUTH'),
    (dir_regex(r'E\.?'), 'EAST'),
    (dir_regex(r'W\.?'), 'WEST')
]

ALIASES = [
    ('ADAM CLAYTON POWELL( JR)?( (BLVD|BOULEVARD))?', 'ADAM CLAYTON POWELL JR BOULEVARD'),
    ('AVENUE OF( THE)? AMERICAS', 'AVENUE OF THE AMERICAS'),
    (r'COLLEGE PT\.? (BLVD|BOULEVARD)', 'COLLEGE POINT BOULEVARD'),
    ('CO-OP CITY', 'COOP CITY')
]

REMOVE = [
    ('(BKLYN|BROOKLYN|QUEENS|BRONX|MANHATTAN|NEW YORK|NYC|SI)$', ''),
    ('(BENSONHURST|CORONA)$', ''),
    (r'\(.+\)$', '')
]

REGEX_REPLACEMENTS = STREETS + DIRECTIONS + REMOVE

# Str, Str -> Lambda
def replace_func(pattern, replacement):
    return lambda s: re.sub(pattern, replacement, s)

ALIASES_FUNCS = list(map(lambda x: replace_func(*x), ALIASES))
REGEX_FUNCS = list(map(lambda x: replace_func(*x), REGEX_REPLACEMENTS))

WORD_NUMBERS = ['ZEROTH', 'FIRST', 'SECOND', 'THIRD',
                'FOURTH', 'FIFTH', 'SIXTH', 'SEVENTH',
                'EIGHTH', 'NINTH', 'TENTH']
SUFFIXES = {
    '0': 'TH',
    '1': 'ST',
    '2': 'ND',
    '3': 'RD',
    '4': 'TH',
    '5': 'TH',
    '6': 'TH',
    '7': 'TH',
    '8': 'TH',
    '9': 'TH'
}


def format_number(matchobj):
    n = matchobj.group('number')
    rest = matchobj.group('rest')
    if int(n) < 11:
        return WORD_NUMBERS[int(n)] + rest
    else:
        tens_digit = str(n)[-2:-1]

        if tens_digit == '1':
            return str(n) + 'TH' + rest
        else:
            return n + SUFFIXES[n[-1]] + rest


def replace_number(string):
    return re.sub(r"(^|(?<=[ ]))(?P<number>\d+)(TH|ST|ND|RD)?(?P<rest>(\b|[ ]).*)", format_number, string)


HOLY_SAINTS = ['JOSEPH', 'MARKS', 'LAWRENCE', 'JAMES',
               'NICHOLAS', 'HOLLIS', 'JOHNS', "JOHN's"]

SAINTS_REGEX = r"ST\.?[ ](?P<street_name>({}))".format('|'.join(HOLY_SAINTS))

def saints(s):
    repl = lambda matchobj: "SAINT " + matchobj.group('street_name')
    return re.sub(SAINTS_REGEX, repl, s)

STREET_FUNCS = ALIASES_FUNCS + [replace_number, saints] + REGEX_FUNCS

# list(of functions), str -> str
def func_chain(funcs, val):
    if len(funcs) == 0:
        return val
    else:
        return func_chain(funcs[1:], funcs[0](val))

def remove_extra_spaces(s):
    return ' '.join([x for x in s.split(' ') if x != ''])


def prepare(s):
    return remove_extra_spaces(s).strip().upper().replace('"', '').replace('-', '')


def normalize_street(street):
    if street is None:
        return None

    s = prepare(street)

    if s == '':
        return None

    return func_chain(STREET_FUNCS, s).replace('.', '').strip()


# remove dashes or spaces...sorry Queens!
def normalize_street_number(number):
    if number is None or number == '':
        return None
    return re.sub(r'(?<=\d)(-|[ ])(?=\d)', '', number).replace('-', '').strip()


APT_STRINGS_TO_REMOVE = ['.', '_', '#', '{', '}', '/']

def clean_apt_str(s):
    s = prepare(s)
    for char_to_remove in APT_STRINGS_TO_REMOVE:
        s = s.replace(char_to_remove, '')
    return s


APT_NUM = r'(?<=\d)(ST|TH|ND|RD)'
# "12F" becomes 12F
# "12TH F" becomes 12FLOOR
EDGE_CASE_FLOOR = APT_NUM + '[ ]F$'
FLOOR_REGEX = r'(?<=\d)[ ]?((FL(OOR|O|R)?)|FW)$'

def normalize_apartment(string):
    if string is None:
        return None

    s = clean_apt_str(string)

    if s == '':
        return None

    s = re.sub(EDGE_CASE_FLOOR, 'FLOOR', s)
    s = re.sub(APT_NUM, '', s)
    s = re.sub(FLOOR_REGEX, 'FLOOR', s)

    return s.replace(' ', '')



import datetime
import logging
import itertools
import os
import subprocess
from tqdm import tqdm

from . import verify
from . import dataset_transformations
from . import sql
from .database import Database
from .typecast import Typecast
from .file import File
from .utility import list_wrap, merge
from .datasets import datasets

BATCH_SIZE = 1000


class Dataset:
    """
    Most CLI actions correspond to a method in this class.
    It is initialized with the name of a dataset:

        hpd_violations = Dataset('hpd_violations')

    To download the files:

        hpd_violations.download_files()

    To load the files into postgres:

        hpd_violations.db_import()
    """

    def __init__(self, dataset_name, args=None):
        self.name = dataset_name
        self.args = args
        self.db = None

        if self.args:
            self.root_dir = self.args.root_dir
        else:
            self.root_dir = './data'

        self.dataset = datasets()[dataset_name]
        self.files = self._files()
        self.schemas = list_wrap(self.dataset['schema'])

    def _files(self):
        return [File(file_dict, folder=self.name, root_dir=self.root_dir) for file_dict in self.dataset['files']]


    def download_files(self):
        """
        Downloads all files for the dataset.

        See ./file.py for more details.
        """
        for f in self.files:
            f.download()


    def db_import(self):
        """
        Inserts the dataset in the postgres.
        Output:  True | Throws
        """
        self.setup_db()
        self.create_schema()

        for schema in self.schemas:
            self.import_schema(schema)

        self.sql_files()

    def index(self):
        """
        Some datasets contains additional indices (notably, full-text-search indices)
        that are not automatically created.

        This method creates those indices.
        It does nothing if the dataset has no additional indexes
        """
        if 'index' in self.dataset:
            for sql_file in self.dataset['index']:
                self.db.execute_sql_file(sql_file)
        else:
            logging.debug('no index files exist for this dataset')

    def transform(self, schema):
        """
        Calls the function in dataset_transformation with the same name
        as the schema.

        If no function exists with the same name as the schema table, it tries
        to call the function with the same name as the dataset

        Input: dict
        Output: generator
        """
        tc = Typecast(schema)

        try:
            rows = getattr(dataset_transformations, schema['table_name'])(self)
        except AttributeError:
            rows = getattr(dataset_transformations, self.name)(self, schema)

        return tc.cast_rows(rows)

    def import_schema(self, schema):
        """
        Imports the schema (table) into postgres in batches.
        """
        rows = self.transform(schema)

        pbar = tqdm(unit='rows')
        while True:
            batch = list(itertools.islice(rows, 0, BATCH_SIZE))
            if len(batch) == 0:
                break
            else:
                pbar.update(len(batch))
                self.db.insert_rows(batch, table_name=schema['table_name'])
        pbar.close()

    def create_schema(self):
        """
        Issues CREATE TABLE statements for all tables in the dataset.
        """
        create_table = lambda name, fields: self.db.sql(sql.create_table(name, fields))

        for s in self.schemas:
            create_table(s['table_name'], s['fields'])

    def sql_files(self):
        """
        Executes all sql files for the dataset.
        """
        if 'sql' in self.dataset:
            for f in self.dataset['sql']:
                self.db.execute_sql_file(f)

    def setup_db(self):
        """
        Establishes the Database object. Used to lazy-load self.db.
        """
        if self.db is None:
            self.db = Database(self.args, table_name=self.name)

    def verify(self):
        """
        Verifies if the tables for this dataset exists and
        if it contains approximately the right number of rows.
        """
        self.setup_db()
        verify.check_dataset(self.db, self.name)

    def dump(self):
        """
        Creates .sql dump file of the datasets.
        Saves the file with the format [DATASET_NAME]-DATE.sql
        """
        tables = ['--table={}'.format(s['table_name']) for s in self.schemas]
        file_arg = '--file=./{}-{}.sql'.format(self.name, datetime.date.today().isoformat())
        cmd = ["pg_dump", "--no-owner", "--clean", "--if-exists", "-w"] + tables + [file_arg]
        subprocess.run(cmd, env=self.pg_env(), check=True)

    def pg_env(self):
        """
        Returns a copy of the environment with postgres environment variables set
        to be the same as the values provided by the args, which are typically given
        on the command line.
        """
        return merge(
            os.environ.copy(),
            {
                'PGHOST': self.args.host,
                'PGPORT': self.args.port,
                'PGUSER': self.args.user,
                'PGDATABASE': self.args.database,
                'PGPASSWORD': self.args.password
            })



import os

from functools import lru_cache
from pathlib import Path
from .utility import read_yml

@lru_cache()
def datasets():
    """
    Returns a dictionary with all defined datasets.
    """
    dataset_dictionary = {}

    for yaml_file in Path(os.path.dirname(__file__)).absolute().glob('./datasets/*.yml'):
        dataset_dictionary[yaml_file.stem] = read_yml(yaml_file)

    return dataset_dictionary



import os
from pathlib import Path
import yaml


def read_yml(file):
    """Reads a yaml file and outputs a Dictionary"""
    with open(file, 'r') as yaml_file:
        return yaml.load(yaml_file, Loader=yaml.FullLoader)


def mkdir(file_path):
    """ Creates all parent directories"""
    Path(os.path.dirname(file_path)).mkdir(parents=True, exist_ok=True)


def list_wrap(x):
    """
    Returns input if is an tuple or list,
    otherwise it wraps x in an list
    """
    if isinstance(x, list) or isinstance(x, tuple):
        return x
    else:
        return [x]


def merge(x, y):
    """Given two dicts, merge them into a new dict as a shallow copy."""
    z = x.copy()
    z.update(y)
    return z



TABLES = {
    'pluto_15v1': {'pluto_15v1': 800_000},
    'pluto_16v2': {'pluto_16v2': 800_000},
    'pluto_17v1': {'pluto_17v1': 800_000},
    'pluto_18v1': {'pluto_18v1': 800_000},
    'pluto_18v2': {'pluto_18v2': 800_000},
    'pluto_19v1': {'pluto_19v1': 800_000},
    'pluto_19v2': {'pluto_19v2': 800_000},
    'dobjobs': {'dobjobs': 1_000_000},
    'dob_violations': {'dob_violations': 2_000_000},
    'ecb_violations': {'ecb_violations': 1_300_000},
    'hpd_violations': {'hpd_violations': 4_000_000},
    'hpd_registrations': {
        'hpd_registrations': 150_000,
        'hpd_contacts': 600_000,
        'hpd_corporate_owners': 30_000,
        'hpd_business_addresses': 130_000,
        'hpd_registrations_grouped_by_bbl': 130_000,
        'hpd_registrations_grouped_by_bbl_with_contacts': 130_000
    },
    'dof_sales': {'dof_sales': 75_000},
    'rentstab': {'rentstab': 45_000},
    'rentstab_v2': {'rentstab_v2': 40_000},
    'rentstab_summary': {'rentstab_summary': 45_000},
    'dob_complaints': {'dob_complaints': 1_000_000},
    'hpd_complaints': {
        'hpd_complaints': 1_000_000,
        'hpd_complaint_problems': 2_000_000
    },
    'pad': {'pad_adr': 1000000},
    'acris': {
        'real_property_remarks': 1_000_000,
        'real_property_legals': 1_000_000,
        'real_property_master': 1_000_000,
        'real_property_parties': 1_000_000,
        'real_property_references': 1_000_000,
        'personal_property_legals': 1_000_000,
        'personal_property_master': 1_000_000,
        'personal_property_parties': 1_000_000,
        'personal_property_references': 1_000_000,
        'personal_property_remarks': 400_000,
        'acris_country_codes': 250,
        'acris_document_control_codes': 123,
        'acris_property_type_codes': 46,
        'acris_ucc_collateral_codes': 8
    },
    'marshal_evictions': {
        'marshal_evictions_17': 17_000,
        'marshal_evictions_18': 19_000,
        'marshal_evictions_19': 16_000,
    },
    'oath_hearings': {'oath_hearings': 10_000_000},
    'hpd_vacateorders': {'hpd_vacateorders': 4_000},
    'j51_exemptions': {'j51_exemptions': 4_000_000}
}


class colors:
    BLUE = '\033[94m'
    GREEN = '\033[92m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'


def check_dataset(db, dataset):
    """ input: nycdb.Database, str """

    for table_name, min_row_count in TABLES[dataset].items():
        if db.table_exists(table_name):
            cnt = db.row_count(table_name)
            if cnt >= min_row_count:
                print(colors.GREEN + table_name + ' has ' + format(cnt, ',') + ' rows' + colors.ENDC)
            else:
                if cnt == 0:
                    print(colors.FAIL + table_name + ' has no rows!' + colors.ENDC)
                else:
                    has_rows = colors.FAIL + table_name + ' has ' + format(cnt, ',') + ' rows.' + colors.ENDC
                    expecting = colors.FAIL + 'Expecting at least ' + colors.BLUE + \
                        format(min_row_count, ',') + colors.ENDC + colors.FAIL + ' rows' + colors.ENDC
                    print(has_rows + expecting)
        else:
            print(colors.FAIL + table_name + ' is missing!' + colors.ENDC)



import codecs
import os
import requests
import logging
from pathlib import Path
from tqdm import tqdm
from .utility import mkdir


class DownloadFailedException(Exception):
    pass


open_str_kwargs = {'mode': 'w', 'encoding': 'utf-8', 'errors': 'replace'}
open_byte_kwargs = {'mode': 'wb'}


def is_csv(dest):
    return dest[-4:].lower() == '.csv'


def open_kwargs(dest):
    if is_csv(dest):
        return open_str_kwargs
    else:
        return open_byte_kwargs


def safe_int(value):
    """
    Attempt to convert the given value to an integer,
    returning None on failure.
    """

    try:
        return int(value)
    except (TypeError, ValueError):
        return None


def download_file(url, dest):
    """
    Downloads a url and saves the result to the destination path

    It will creates parent directory of the destination path,
    if they they don't exist.

    If the destination file exists and is not empty, it assumes the file has
    already been downloaded and will skip downloading the file accordingly.
    """
    mkdir(dest)

    if Path(dest).exists() and os.stat(dest).st_size > 0:
        logging.info("{} has already been downloaded, skipping".format(url))
        return True

    try:
        logging.info("Downloading {url} to {dest}".format(url=url, dest=dest))
        is_csv_file = is_csv(dest)
        r = requests.get(url, stream=True)
        total_bytes = safe_int(r.headers.get('content-length'))
        pbar = tqdm(total=total_bytes, unit='B', unit_scale=True, unit_divisor=1024)
        with open(dest, **open_kwargs(dest)) as f:
            for chunk in r.iter_content(chunk_size=(512 * 1024)):
                if chunk:
                    pbar.update(len(chunk))
                    if is_csv_file:
                        f.write(codecs.decode(chunk, encoding='utf-8', errors='replace'))
                    else:
                        f.write(chunk)
        pbar.close()
        return True
    except:
        raise DownloadFailedException("Could not download: {}".format(url))


class File:
    """Wrapper around a file"""

    def __init__(self, file_dict, root_dir='./data', folder=''):
        self.root_dir = root_dir
        self.url = file_dict['url']
        self.dest = self._dest(file_dict)

    def download(self):
        download_file(self.url, self.dest)
        return self

    def _dest(self, file_dict):
        if 'dest' in file_dict:
            file_path = file_dict['dest']
        else:
            file_path = file_dict['url'].split('/')[-1]
        return os.path.abspath(os.path.join(self.root_dir, file_path))



"""
Typecasting for NYCDB

All values are converted into a suitable python class before
being passed to psycopg2 to be inserted into postgres.
See http://initd.org/psycopg/docs/usage.html#adaptation-of-python-values-to-sql-types
for how psycopg2 converts python types into postgres types
"""

import copy
import re
import datetime
from decimal import Decimal, InvalidOperation

YES_VALUES = [1, True, 'T', 't', 'true', 'True', 'TRUE', '1', 'y', 'Y', "YES", 'Yes', 'x', 'X']
NO_VALUES = ['0', 0, False, 'False', 'f', 'F', 'false', 'FALSE', 'N', 'n', 'NO', 'No', 'no']
INTEGER_TYPES = ['integer', 'smallint', 'bigint', 'int']


def downcase_fields_and_values(d):
    """downcase keys and values in dictionary"""
    return dict((k.lower(), v.strip().lower()) for k, v in d.items())


def integer(i):
    if isinstance(i, int):
        return i
    try:
        int_str = i.strip().replace('$', '')

        if int_str == '.' or int_str == '':
            return None
        elif '.' in i:
            return int(int_str.split('.')[0])
        else:
            return int(int_str)

    except ValueError:
        return None


def text(x):
    if x is None:
        return None
    s = str(x).strip()
    if s == '':
        return None
    else:
        return s


def char(x, n):
    if x is None:
        return None
    val = str(x)
    if len(val) > n:
        return val.strip()[0:n]
    else:
        return val


def numeric(x):
    try:
        return Decimal(x)
    except (InvalidOperation, TypeError):
        return None


def to_float(x):
    if x is None:
        return None
    
    if isinstance(x, float):
        return x

    try:
        return float(x)
    except ValueError:
        return None

def mm_dd_yyyy(date_str):
    try:
        month, day, year = map(int, date_str[0:10].split('/'))
        return datetime.date(year, month, day)
    except ValueError:
        return None


# TODO: allow for different date inputs besides mm/dd/yyyy
#  03/04/2015 12:00:00 AM
def date(x):
    if isinstance(x, (datetime.date, datetime.datetime)):
        return x
    # checks for 2018-12-31 date input
    if re.match(r'\d{4}-\d{1,2}-\d{1,2}', x):
        try:
            return datetime.datetime.strptime(x, '%Y-%m-%d').date()
        except ValueError:
            return None
    # checks for 20181231 date input
    elif re.match(r'[0-9]{8}', x):
        try:
            return datetime.datetime.strptime(x, '%Y%m%d').date()
        except ValueError:
            return None
    # checks for 12/31/2018 date input
    elif len(x) == 10 and len(x.split('/')) == 3:
        return mm_dd_yyyy(x)
    # checks for 12/31/2018 12:00:00 AM date input
    elif len(x) == 22 and len(x[0:10].split('/')) == 3:
        return mm_dd_yyyy(x)
    else:
        return None


def time(x):
    """
    Converts string into datetime.time
    Example input: 13:01:00
    """
    if isinstance(x, datetime.time):
        return x
    if isinstance(x, str) and re.match(r'^\d{2}:\d{2}:\d{2}$', x.strip()):
        try:
            return datetime.time(*map(int, x.strip().split(':')))
        except ValueError:
            return None


def boolean(x):
    if x in YES_VALUES:
        return True
    elif x in NO_VALUES:
        return False
    else:
        return None


def text_array(x, sep=","):
    return x.strip().split(sep)


def char_cast(n):
    n = copy.copy(n)

    def to_char(x):
        return char(x, n)

    return to_char


class Typecast():
    def __init__(self, schema):
        self.fields = downcase_fields_and_values(schema['fields'])
        self.cast = self.generate_cast()

    def cast_rows(self, rows):
        """
        input: Iterable
        output: Iterable
        """
        for row in rows:
            yield self.cast_row(row)

    def cast_row(self, row):
        """
        Converts values of dictionary by type of dataset
        input: Dict
        output: Dict
        """
        try:
            d = {}
            for column, val in row.items():
                d[column] = self.cast[column.lower()](val)
            return d
        except:
            # print the row for debugging:
            print(row)
            raise

    def generate_cast(self):
        """
        Generates conversation table for dataset schema
        """
        d = {}
        for k, v in self.fields.items():
            if 'serial' in v:
                continue
            elif v[0:4] == 'char':
                n = int(re.match(r'char\((\d+)\)', v).group(1))
                d[k] = char_cast(n)
            elif v in INTEGER_TYPES:
                d[k] = lambda x: integer(x)
            elif v == 'text':
                d[k] = lambda x: text(x)
            elif v == 'boolean':
                d[k] = lambda x: boolean(x)
            elif v == 'date':
                d[k] = lambda x: date(x)
            elif 'time' in v:
                d[k] = lambda x: time(x)
            elif v in ['real', 'double precision']:
                d[k] = lambda x: to_float(x)
            elif v == 'numeric':
                d[k] = lambda x: numeric(x)
            elif v == 'text[]':
                d[k] = lambda x: text_array(x)
            else:
                d[k] = lambda x: x
        return d



import pyproj

ny_state_plane = pyproj.Proj("epsg:2263", preserve_units=True)
wgs84 = pyproj.Proj(init="epsg:4326")
transformer = pyproj.Transformer.from_proj(ny_state_plane, wgs84)

def ny_state_coords_to_lat_lng(xcoord, ycoord):
    return transformer.transform(xcoord, ycoord)



"""nycdb - nyc housing database"""

__version__ = '0.1.0'
__author__ = 'ziggy <ziggy@elephant-bird.net>'
#__all__ = []

from . import typecast, sql, transform, bbl, dof_parser, address, geo
from .database import Database
from .dataset import Dataset
from .datasets import datasets
from .file import File


#datasets = _datasets()



import os
import psycopg2
import psycopg2.extras
from . import sql


class Database:
    """
    Database connection. This is a wrapper around psycopg2.connection, accessible at `.conn`
    see http://initd.org/psycopg/docs/connection.html for psycopg2's documentation


    """
    def __init__(self, args, table_name=None):
        """
        args is a Namespace that must contain: user, password, database, host, and port
        """
        self.conn = psycopg2.connect(
            user=args.user,
            password=args.password,
            host=args.host,
            database=args.database,
            port=args.port
        )

        self.table_name = table_name

        self.connection_params = {
            'user': args.user,
            'password': args.password,
            'host': args.host,
            'database': args.database,
            'port': args.port
        }

    def sql(self, SQL):
        """ executes single sql statement """
        with self.conn.cursor() as curs:
            curs.execute(SQL)
        self.conn.commit()

    def insert_rows(self, rows, table_name=None):
        """
        Inserts many rows, all in the same transaction, using psycopg2.extras.execute_values
        """

        if table_name is None:
            table_name = self.table_name

        with self.conn.cursor() as curs:
            sql_str, template = sql.insert_many(table_name, rows)
            try:
                psycopg2.extras.execute_values(
                    curs,
                    sql_str,
                    rows,
                    template=template,
                    page_size=len(rows)
                )
            except psycopg2.DataError:
                print(rows) # useful for debugging
                raise
        self.conn.commit()

    def execute_sql_file(self, sql_file):
        """
        Executes the provided sql file.
        It assumes the path is relative to ./sql
        """
        file_path = os.path.join(os.path.dirname(__file__), 'sql', sql_file)

        with open(file_path, 'r', encoding='utf-8') as f:
            self.sql(f.read())

    def execute_and_fetchone(self, query):
        """
        Execute the given query and returns the first row
        """
        with self.conn.cursor() as curs:
            curs.execute(query)
            return curs.fetchone()[0]


    def table_exists(self, table_name):
        """Tests if the table exists"""
        query = "SELECT EXISTS(SELECT 1 FROM information_schema.tables where table_name = '{0}')".format(table_name)
        return self.execute_and_fetchone(query)

    def row_count(self, table_name):
        """returns the row count of the table"""
        query = "SELECT COUNT(*) from {0}".format(table_name)
        return self.execute_and_fetchone(query)

    def password_file_contents(self):
        return "{host}:{port}:{database}:{user}:{password}".format(**self.connection_params)
